#' Find marker genes using a edgeR.
#'
#' Wraps functions from the edgeR package.
#'
#' @param sce A SingleCellExperiment object
#' @param pars parameters passed to the function
#'
#' @return A list containing the time of the computation and results of the
#'  method.
#'
run_edger <- function(sce, pars) {

  `%>%` <- dplyr::`%>%`

  # TODO: Extend the pairwise case.

  times <- system.time({

    n_groups <- length(unique(colLabels(sce)))
    out <- vector("list", length = n_groups)

    for (j in seq_len(n_groups)) {

      group <- levels(colLabels(sce))[[j]]
      x <- factor(as.numeric(colLabels(sce) == group))

      # Fit the edgeR model.
      dge <- edgeR::DGEList(counts(sce), group = x)

      # FIXME:
      # 1) This warns; why?
      # 2) Should we calculate normalizing factors here or use the existing
      #    factors from i.e. scran?
      dge <- edgeR::calcNormFactors(dge)

      if (pars$covariate == "cdr") {
        # Use cellular detection rate as an additional covariate.
        cdr <- scale(colMeans(counts(sce) > 0))
        f <- ~ cdr + x
      } else if (pars$covariate == "none") {
        f <- ~ x
      } else {
        stop("Invalid covariate specification.", call. = FALSE)
      }

      design <- model.matrix(f)

      dge <- edgeR::estimateDisp(dge, design = design)

      if (pars$fit_method == "ql") {
        fit <- edgeR::glmQLFit(dge, design = design)
        stat <- edgeR::glmQLFTest(fit)
      } else if (pars$fit_method == "glm") {
        fit <- edgeR::glmFit(dge, design = design)
        stat <- edgeR::glmLRT(fit)
      } else {
        stop("Invalid `fit_method`.", call. = FALSE)
      }
      tt <- edgeR::topTags(stat, n = Inf)

      # Process the returned list of tags.
      group_out <- as.data.frame(tt) %>%
        tibble::rownames_to_column("gene") %>%
        dplyr::rename(
          log_fc = logFC,
          p_value = PValue,
          p_value_adj = FDR,
        ) %>%
        dplyr::select(
          -c(logCPM)
        ) %>%
        dplyr::mutate(cluster = group, raw_statistic = 0) %>%
        tibble::as_tibble()

      if (pars$fit_method == "ql") {
        group_out <- group_out %>%
          dplyr::mutate(scaled_statistic = sqrt(F)) %>%
          dplyr::select(-F)
      } else if (pars$fit_method == "glm") {
        group_out <- group_out %>%
          # FIXME: Does this make sense?
          dplyr::mutate(scaled_statistic = sqrt(LR)) %>%
          dplyr::select(-LR)
      } else {
        stop("Invalid `fit_method`.", call. = FALSE)
      }

      out[[j]] <- group_out
    }

    result <- dplyr::bind_rows(!!!out)
  })

  # Get the elapsed time.
  time <- times[[3]]

  validate_mgs_result(result)

  list(time = time, result = result, raw_result = result, pars = pars)
}
