#' Find marker genes using scanpy
#'
#' @param sce A SingleCellExperiment object,
#' @param pars A list of parameters passed to the function.
#'
#' @return A list containing the time of the computation and results of the
#'  method.
#'
run_scanpy <- function(sce, pars) {

  # Ensure we use the correct conda environment.
  # FIXME: Move to .Rprofile.
  Sys.setenv(RETICULATE_PYTHON = "/home/jpullin/.conda/envs/mage/bin/python")
  library(reticulate)

  # Load scanpy (R side).
  sc <- import("scanpy")

  # Clear {splatter} `Params` from metadata which cannot be converted.
  sce@metadata <- list()

  # Convert the SingleCellExperiment object to an Annotated Data object.
  # We do not use {zellkonverter} as it didn't seem to work when tested.
  adata <- sc$AnnData(
    X   = t(logcounts(sce)),
    obs = as.data.frame(colData(sce)),
    var = as.data.frame(rowData(sce))
  )

  # Deal with the issue of hyphens in the method names and `tie_correct`.
  tie_correct <- FALSE
  if (pars$test_use == "t") {
    test_use <- "t-test"
  } else if (pars$test_use == "tover") {
    test_use <- "t-test_overestim_var"
  } else if (pars$test_use == "wilcoxontiecorrect") {
    test_use <- "wilcoxon"
    tie_correct <- TRUE
  } else {
    test_use <- pars$test_use
  }

  # Find the marker genes.
  times <- system.time({
    sc$tools$rank_genes_groups(
      adata,
      groupby = "label",
      method = test_use,
      tie_correct = tie_correct,
      rankby_abs = as.logical(pars$rankby_abs)
    )
  })

  py$adata <- adata

  print(adata)

  # We use the 'elapsed' time.
  time <- times[[3]]

  # The hard part: extracting the marker genes.

  # Load pandas in the main python module.
  py_run_string("import pandas as pd")

  name_lookup <- list(
    # FIXME: Not all are strictly t-values...
    names = "gene",
    scores = "scaled_statistic",
    pvals = "p_value",
    pvals_adj = "p_value_adj",
    logfoldchanges = "log_fc"
  )

  out <- list()
  for (ele in names(name_lookup)) {

    # We have to do this on the 'python side' to prevent array type 20 errors.
    py_run_string(
      glue::glue("{ele} = pd.DataFrame(adata.uns['rank_genes_groups']['{ele}'])")
    )
    out <- c(out,
      list(tidyr::pivot_longer(
        py[[ele]],
        cols = everything(),
        names_to = "cluster",
        values_to = name_lookup[[ele]]
      ))
    )
  }

  cluster <- tibble::tibble(cluster = out[[1]]$cluster)
  out <- lapply(out, function(x) dplyr::select(x, -cluster))

  result <- dplyr::bind_cols(cluster, !!!out)
  result <- dplyr::mutate(result, raw_statistic = 0)

  # In the past the output from Scanpy grouped the results for each cluster
  # together with genes ranked. At some point however this changed for an
  # unknown reason, so we rectify this here. We DO NOT change the ranking of
  # scanpy.
  result <- dplyr::arrange(result, cluster)

  # FIXME: Can we save the raw result in the anndata object?
  list(time = time, result = result, raw_result = result, pars = pars)
}
